﻿using Microsoft.Build.Framework;
using Microsoft.Build.Utilities;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using devtm.Aop.Convert;
using devtm.Aop;


namespace devtm.Aop.Task
{

    public class AopTask : Microsoft.Build.Utilities.Task, ILogging
    {


        public override bool Execute()
        {
            this.Log.LogMessage("FileName : " + FileName);

            // Converti par des attributs declarés sur les methodes.
            MethodValidator validator = new MethodValidator();

            // Traitement
            CodeDomIILVisitorConverter visitor = new CodeDomIILVisitorConverter(FileName, this, true);


            try
            {
                visitor.Analyze(validator, FileName);
            }
            catch (Exception ex)
            {
                this.Log.LogError(ex.Message);
                throw;
            }


            return true;
        }

        [Required]
        public string FileName { get; set; }

        [Output]
        public int ExitCode { get; set; }




        #region Logs
        
        public void LogError(string message, params object[] messageArgs)
        {
            this.Log.LogError(message, messageArgs);
        }

        public void LogErrorFromException(Exception exception)
        {
            this.Log.LogErrorFromException(exception);
        }

        public void LogErrorFromException(Exception exception, bool showStackTrace)
        {
            this.Log.LogErrorFromException(exception, showStackTrace);
        }

        public void LogErrorFromException(Exception exception, bool showStackTrace, bool showDetail, string file)
        {
            this.Log.LogErrorFromException(exception, showStackTrace, showDetail, file);
        }

        public void LogErrorFromResources(string messageResourceName, params object[] messageArgs)
        {
            this.Log.LogErrorFromResources(messageResourceName, messageArgs);
        }

        public void LogErrorFromResources(string subcategoryResourceName, string errorCode, string helpKeyword, string file, int lineNumber, int columnNumber, int endLineNumber, int endColumnNumber, string messageResourceName, params object[] messageArgs)
        {
            this.Log.LogErrorFromResources(subcategoryResourceName, errorCode, helpKeyword, file, lineNumber, columnNumber, endLineNumber, endColumnNumber, messageResourceName, messageArgs);
        }

        public void LogErrorWithCodeFromResources(string messageResourceName, params object[] messageArgs)
        {
            this.Log.LogErrorWithCodeFromResources(messageResourceName, messageArgs);
        }

        public void LogErrorWithCodeFromResources(string subcategoryResourceName, string file, int lineNumber, int columnNumber, int endLineNumber, int endColumnNumber, string messageResourceName, params object[] messageArgs)
        {
            this.Log.LogErrorWithCodeFromResources(subcategoryResourceName, file, lineNumber, columnNumber, endLineNumber, endColumnNumber, messageResourceName, messageArgs);
        }

        public void LogExternalProjectFinished(string message, string helpKeyword, string projectFile, bool succeeded)
        {
            this.Log.LogExternalProjectFinished(message, helpKeyword, projectFile, succeeded);
        }

        public void LogExternalProjectStarted(string message, string helpKeyword, string projectFile, string targetNames)
        {
            this.Log.LogExternalProjectStarted(message, helpKeyword, projectFile, targetNames);
        }

        public void LogMessage(string message, params object[] messageArgs)
        {
            this.Log.LogMessage(message, messageArgs);
        }

        public void LogMessageFromResources(string messageResourceName, params object[] messageArgs)
        {
            this.Log.LogMessageFromResources(messageResourceName, messageArgs);
        }

        public bool LogMessagesFromFile(string fileName)
        {
            return this.Log.LogMessagesFromFile(fileName);
        }

        public void LogWarning(string message, params object[] messageArgs)
        {
            this.Log.LogWarning(message, messageArgs);
        }

        public void LogWarning(string subcategory, string warningCode, string helpKeyword, string file, int lineNumber, int columnNumber, int endLineNumber, int endColumnNumber, string message, params object[] messageArgs)
        {
            this.Log.LogWarning(subcategory, warningCode, helpKeyword, file, lineNumber, columnNumber, endLineNumber, endColumnNumber, message, messageArgs);
        }

        public void LogWarningFromException(Exception exception)
        {
            this.Log.LogWarningFromException(exception);
        }

        public void LogWarningFromException(Exception exception, bool showStackTrace)
        {
            this.Log.LogWarningFromException(exception, showStackTrace);
        }

        public void LogWarningFromResources(string messageResourceName, params object[] messageArgs)
        {
            this.Log.LogWarningFromResources(messageResourceName, messageArgs);
        }

        public void LogWarningFromResources(string subcategoryResourceName, string warningCode, string helpKeyword, string file, int lineNumber, int columnNumber, int endLineNumber, int endColumnNumber, string messageResourceName, params object[] messageArgs)
        {
            this.Log.LogWarningFromResources(subcategoryResourceName, warningCode, helpKeyword, file, lineNumber, columnNumber, endLineNumber, endColumnNumber, messageResourceName, messageArgs);
        }

        public void LogWarningWithCodeFromResources(string messageResourceName, params object[] messageArgs)
        {
            this.Log.LogWarningWithCodeFromResources(messageResourceName, messageArgs);
        }

        public void LogWarningWithCodeFromResources(string subcategoryResourceName, string file, int lineNumber, int columnNumber, int endLineNumber, int endColumnNumber, string messageResourceName, params object[] messageArgs)
        {
            this.Log.LogWarningWithCodeFromResources(subcategoryResourceName, file, lineNumber, columnNumber, endLineNumber, endColumnNumber, messageResourceName, messageArgs);
        }

        #endregion

    }

}