using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel;
using System.ServiceModel.Channels;

namespace System.ServiceModel.Dynamic.Test.ServiceModel.Cache
{
    public class ClientCacheChannelBase<TChannel> : ChannelBase
        where TChannel : class, IChannel
    {

        #region Private Fields
        TChannel innerChannel; 
        #endregion

        #region Protected Constructors
        protected ClientCacheChannelBase(
            ChannelManagerBase manager, TChannel innerChannel)
            : base(manager)
        {
            if (innerChannel == null)
            {
                throw new ArgumentException("ClientCacheChannelBase requires a non-null inner channel.", "innerChannel");
            }

            this.innerChannel = innerChannel;
        }
        
        #endregion

        #region Protected Properties
        public TChannel InnerChannel
        {
            get
            {
                return this.innerChannel;
            }
        } 
        #endregion

        #region Public Methods
        public override T GetProperty<T>()
        {
            T baseProperty = base.GetProperty<T>();
            if (baseProperty != null)
            {
                return baseProperty;
            }

            return this.InnerChannel.GetProperty<T>();
        } 
        #endregion

        #region Protected Methods
        protected override void OnAbort()
        {
            this.innerChannel.Abort();
        }

        protected override IAsyncResult OnBeginClose(TimeSpan timeout, AsyncCallback callback, object state)
        {
            return this.innerChannel.BeginClose(timeout, callback, state);
        }

        protected override IAsyncResult OnBeginOpen(TimeSpan timeout, AsyncCallback callback, object state)
        {
            return this.innerChannel.BeginOpen(timeout, callback, state);
        }

        protected override void OnClose(TimeSpan timeout)
        {
            this.innerChannel.Close(timeout);
        }

        protected override void OnEndClose(IAsyncResult result)
        {
            this.innerChannel.EndClose(result);
        }

        protected override void OnEndOpen(IAsyncResult result)
        {
            this.innerChannel.EndOpen(result);
        }

        protected override void OnOpen(TimeSpan timeout)
        {
            this.innerChannel.Open(timeout);
        } 
        #endregion    
    }
}
