using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel;
using System.ServiceModel.Channels;

namespace System.ServiceModel.Dynamic.Test.ServiceModel.Cache
{
    public class ClientCacheChannelFactory<TChannel> : ChannelFactoryBase<TChannel>
    {

        #region Private Fields
        IChannelFactory<TChannel> innerChannelFactory;
        private TimeSpan timeout;
        private KeyCreationMethod keyCreationMethod;
        private string cacheType;
        private bool enabled;
        private bool header;
        private List<int> indexList = new List<int>();
        private Dictionary<string, OperationSettings> operationDictionary;
        #endregion

        #region Public Constructor
        public ClientCacheChannelFactory(BindingContext context, 
                                         TimeSpan timeout,                                          
                                         KeyCreationMethod keyCreationMethod,
                                         string cacheType,
                                         bool enabled,
                                         bool header,
                                         List<int> indexList,
                                         Dictionary<string, OperationSettings> operationDictionary)
        {
            innerChannelFactory = context.BuildInnerChannelFactory<TChannel>();
            if (innerChannelFactory == null)
            {
                throw new InvalidOperationException("ClientCacheChannelFactory requires an inner IChannelFactory");
            }
            this.timeout = timeout;
            this.keyCreationMethod = keyCreationMethod;
            this.cacheType = cacheType;
            this.enabled = enabled;
            this.header = header;
            this.indexList = indexList;
            this.operationDictionary = operationDictionary;
        } 
        #endregion

        #region Public Methods

        public override T GetProperty<T>()
        {
            T baseProperty = base.GetProperty<T>();
            if (baseProperty != null)
                return baseProperty;

            return this.innerChannelFactory.GetProperty<T>();
        }

        protected override void OnOpen(TimeSpan timeout)
        {
            this.innerChannelFactory.Open(timeout);
        }

        protected override IAsyncResult OnBeginOpen(TimeSpan timeout, AsyncCallback callback, object state)
        {
            return this.innerChannelFactory.BeginOpen(timeout, callback, state);
        }

        protected override void OnEndOpen(IAsyncResult result)
        {
            this.innerChannelFactory.EndOpen(result);
        }

        protected override void OnClose(TimeSpan timeout)
        {
            this.innerChannelFactory.Close(timeout);
        }

        protected override IAsyncResult OnBeginClose(TimeSpan timeout, AsyncCallback callback, object state)
        {
            return this.innerChannelFactory.BeginClose(timeout, callback, state);
        }

        protected override void OnEndClose(IAsyncResult result)
        {
            this.innerChannelFactory.EndClose(result);
        }

        protected override TChannel OnCreateChannel(EndpointAddress to, Uri via)
        {
            TChannel innerChannel = this.innerChannelFactory.CreateChannel(to, via);

            if (typeof(TChannel) == typeof(IRequestChannel) || typeof(TChannel) == typeof(IRequestSessionChannel))
            {
                return (TChannel)(object)new ClientCacheRequestChannel<TChannel>(this,
                                                                                 (IRequestChannel)innerChannel,
                                                                                 timeout,
                                                                                 keyCreationMethod,
                                                                                 cacheType,
                                                                                 enabled,
                                                                                 header,
                                                                                 indexList,
                                                                                 operationDictionary);
            }
            throw new InvalidOperationException();
        } 
        #endregion
    
    }
}

