using System;
using System.Collections.Generic;
using System.ServiceModel;
using System.ServiceModel.Description;
using System.ServiceModel.Channels;
using System.ServiceModel.Dispatcher;
using System.ServiceModel.Dynamic.Test.ServiceModel.Cache.Configuration;

namespace System.ServiceModel.Dynamic.Test.ServiceModel.Cache
{
    /// <summary>
    /// ClientCacheEndpointBehavior implements the IEndpointBehavior interface 
    /// </summary>
    public class ClientCacheEndpointBehavior : IEndpointBehavior
    {
        #region Private Fields
        private TimeSpan timeout;
        private string indexes;
        private KeyCreationMethod keyCreationMethod;
        private string cacheType;
        private bool enabled;
        private bool header;
        private OperationElementCollection operations;
        #endregion

        #region Public Constructors
        public ClientCacheEndpointBehavior(TimeSpan timeout,                                           
                                           string indexes,
                                           KeyCreationMethod keyCreationMethod,
                                           string cacheType,
                                           bool enabled,
                                           bool header,
                                           OperationElementCollection operations)
        {
            this.timeout = timeout;
            this.indexes = indexes;
            this.keyCreationMethod = keyCreationMethod;
            this.cacheType = cacheType;
            this.enabled = enabled;
            this.header = header;
            this.operations = operations;
        }
        #endregion

        #region IEndpointBehavior Members
        public void AddBindingParameters(ServiceEndpoint endpoint, BindingParameterCollection bindingParameters)
        {
            
        }

        public void ApplyClientBehavior(ServiceEndpoint endpoint, ClientRuntime clientRuntime)
        {
            BindingElementCollection collection = endpoint.Binding.CreateBindingElements();
            List<BindingElement> list = new List<BindingElement>();
            list.Add(new ClientCacheBindingElement(this.timeout,                                                    
                                                   this.indexes,
                                                   this.keyCreationMethod, 
                                                   this.cacheType, 
                                                   this.enabled,
                                                   this.header,
                                                   this.operations));
            list.AddRange(collection);
            CustomBinding customBinding = new CustomBinding(list);
            customBinding.CloseTimeout = endpoint.Binding.CloseTimeout;
            customBinding.Name = endpoint.Binding.Name;
            customBinding.Namespace = endpoint.Binding.Namespace;
            customBinding.OpenTimeout = endpoint.Binding.OpenTimeout;
            customBinding.ReceiveTimeout = endpoint.Binding.ReceiveTimeout;
            customBinding.SendTimeout = endpoint.Binding.SendTimeout;
            endpoint.Binding = customBinding;
        }

        public void ApplyDispatchBehavior(ServiceEndpoint endpoint, EndpointDispatcher endpointDispatcher)
        {
            
        }

        public void Validate(ServiceEndpoint endpoint)
        {
            return;
        }
        #endregion
    }
}
