using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml;
using System.Xml.Schema;
using System.Configuration;
using System.Xml.Serialization;
using System.ComponentModel;
using System.ServiceModel.Dynamic.Test.ServiceModel.Cache;

namespace System.ServiceModel.Dynamic.Test.ServiceModel.Cache.Configuration
{


    public partial class ClientCacheConfigSection : System.Configuration.ConfigurationSection, IXmlSerializable
    {


        #region Common


        /// <summary>
        /// Used to initialize a default set of values for the <see cref="ClientCacheConfigSection" /> 
        /// object.
        /// </summary>
        public void InitializeDefaultValue()
        {
            base.InitializeDefault();
        }


        /// <summary>
        /// Indicates whether this configuration element has been modified since it was
        /// last saved or loaded, when implemented in a derived class.
        /// </summary>
        /// <returns><c>true</c> if the element has been modified; otherwise, <c>false</c>.</returns>
        public bool ObjetctIsModified()
        {
            return base.IsModified();
        }


        /// <summary>
        /// Gets a value indicating whether the <see cref="ClientCacheConfigSection" />
        /// object is read-only.
        /// </summary>
        /// <returns><c>true</c> if the MangasCoreConfigSection object is read-only otherwise, <c>false</c>.</returns>
        public bool ObjetctIsReadOnly()
        {
            return base.IsReadOnly();
        }


        #region IXmlSerializable


        /// <summary>
        /// Updates the <see cref="ClientCacheConfigSection" /> with the values from an <see cref="XmlReader" />.													
        /// </summary>													
        /// <param name="reader">The <see cref="XmlReader" /> that reads the configuration source located at the element that describes the configuration section.</param>													
        public void ReadXml(XmlReader reader)
        {
            if (reader == null) throw new ArgumentNullException("reader");

            reader.Read();
            DeserializeSection(reader);
        }


        /// <summary>
        /// Writes the <see cref="ClientCacheConfigSection" /> values as an XML element to an <see cref="XmlWriter" />.													
        /// </summary>													
        /// <param name="writer">The <see cref="XmlWriter" /> that writes to the configuration source.</param>													
        public void WriteXml(XmlWriter writer)
        {
            if (writer == null) throw new ArgumentNullException("writer");

            String serialized = SerializeSection(this, "MangasCoreConfigSection", ConfigurationSaveMode.Full);
            writer.WriteRaw(serialized);
        }


        /// <summary>													
        /// Returns the XML schema for the configuration section.													
        /// </summary>													
        /// <returns>A string with the XML schema, or <see langword="null" /> (<b>Nothing</b> 													
        /// in Visual Basic) if there is no schema.</returns>													
        public XmlSchema GetSchema()
        {
            return null;
        }


        #endregion IXmlSerializable


        /// <summary>
        /// Override the current <see cref="ClientCacheConfigSection" /> object with the parameters from the element.
        /// </summary>
        /// <returns></returns>
        public void OverrideFrom(ClientCacheConfigSection element)
        {

            StringWriter sWriter = new StringWriter();
            using (XmlWriter writer = XmlWriter.Create(sWriter))
            {
                element.WriteXml(writer);
                writer.Flush();
            }

            StringReader sReader = new StringReader(sWriter.ToString());
            using (XmlReader reader = XmlReader.Create(sReader))
            {
                this.ReadXml(reader);
            }

        }

        #endregion Common


        public const string DefaultName = "clientCacheWcf";

        private static ClientCacheConfigSection _configuration;
        public static ClientCacheConfigSection Configuration
        {
            get
            {
                return _configuration ?? (_configuration = System.Configuration.ConfigurationManager.GetSection(DefaultName) as ClientCacheConfigSection);
            }
            set
            {
                _configuration = value;
            }
        }


        /// <summary>
        /// Gets a writable <see cref="ClientCacheConfigSection" /> object with the parameters from the current.
        /// </summary>
        /// <returns></returns>
        public ClientCacheConfigSection AsWritable()
        {

            ClientCacheConfigSection config = new ClientCacheConfigSection();
            StringWriter sWriter = new StringWriter();
            using (XmlWriter writer = XmlWriter.Create(sWriter))
            {
                this.WriteXml(writer);
                writer.Flush();
            }

            StringReader sReader = new StringReader(sWriter.ToString());
            using (XmlReader reader = XmlReader.Create(sReader))
            {
                config.ReadXml(reader);
            }

            return config;

        }


        #region Refs


        #region OperationElementCollection

        /// <summary>
        /// Get the global cryptography parameters
        /// </summary>
        /// <returns>		
        /// Objet <see cref="Mangas.Core.Configuration.CryptographyElement" />.");
        /// </returns>
        [DisplayName("ConfigurationSetting")]
        [Description("Get the global operations settings parameters for the client cache")]
        [Category("Misc")]
        [ConfigurationProperty(OperationSettingsConst, IsRequired = false)]
        [ConfigurationCollection(typeof(OperationElement), AddItemName = "add", ClearItemsName = "clear", RemoveItemName = "del")]
        public OperationElementCollection OperationSettings
        {
            get { return (OperationElementCollection)this[OperationSettingsConst]; }
        }
        public const string OperationSettingsConst = "operationSettings";

        #endregion


        #endregion Refs

    }


	  
}
