using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel.Configuration;
using System.Configuration;

namespace System.ServiceModel.Dynamic.Test.ServiceModel.Cache
{
    public class OperationSettings
    {
        #region Private Constants
        
        #endregion

        #region Private Fields

        private System.ServiceModel.Dynamic.Test.ServiceModel.Cache.ICache _cache;

        #endregion
        
        #region Public Constructors


        public OperationSettings()
        {
            this.KeyCreationMethod = KeyCreationMethod.Action;
            this.CacheType = System.ServiceModel.Dynamic.Test.ServiceModel.Cache.CacheFactory.DefaultCache;
            this.Timeout = TimeSpan.FromMinutes(5);            
            this.Enabled = true;
            this.Header = false;
        }

        public OperationSettings(string indexes) 
        {
            this.KeyCreationMethod = KeyCreationMethod.Action;
            this.CacheType = System.ServiceModel.Dynamic.Test.ServiceModel.Cache.CacheFactory.DefaultCache;
            this.Timeout = TimeSpan.FromMinutes(5);
            this.Enabled = true;
            this.Header = false;

            if (!string.IsNullOrEmpty(indexes))
            {
                int index;
                string[] items = indexes.Split(',');
                for (int i = 0; i < items.Length; i++)
                {
                    if (int.TryParse(items[i], out index) &&
                        !Indexes.Contains(index))
                    {
                        Indexes.Add(index);
                    }
                }
            }
        }


        public OperationSettings(string indexes,
                                 KeyCreationMethod keyCreationMethod,
                                 string cacheType,
                                 TimeSpan timeout,
                                 bool enabled,
                                 bool header) :this(indexes)
        {
            this.KeyCreationMethod = keyCreationMethod;
            this.CacheType = cacheType;
            this.Timeout = timeout;            
            this.Enabled = false;
            this.Header = header;
        }


        public OperationSettings(List<int> indexList,
                                 KeyCreationMethod keyCreationMethod,
                                 string cacheType,
                                 TimeSpan timeout,                                 
                                 bool enabled,
                                 bool header)
        {
            this.Indexes = indexList;
            this.KeyCreationMethod = keyCreationMethod;
            this.CacheType = cacheType;
            this.Timeout = timeout;
            this.Enabled = enabled;
            this.Header = header;
        }
        #endregion

        #region Public Properties

        /// <summary>
        /// Gets the list of indexes of parameters to be used to calculate the cache key.
        /// This property is used only when the KeyConstructionMethod = Indexed
        /// </summary>
        public List<int> Indexes { get; set; }

        /// <summary>
        /// Gets or sets a value indicating whether the WCF client cache is enabled.
        /// When the value is false, the channel stack always invokes the target service.
        /// </summary>
        public bool Enabled { get; set; }

        /// <summary>
        /// Gets or sets a value indicating whether a custom header is added to the response to 
        /// indicate the source of the WCF message (cache or service).
        /// </summary>
        public bool Header { get; set; }

        /// <summary>
        /// Gets or sets the amount of time the object should reside in the cache before expiration.
        /// </summary>
        public TimeSpan Timeout { get; set; }

        /// <summary>
        /// Gets or sets the method used to calculate the key of the cache item
        /// </summary>
        public KeyCreationMethod KeyCreationMethod { get; set; }

        /// <summary>
        /// Gets or sets the cache type used to store items.
        /// </summary>
        public string CacheType { get; set; }

        public System.ServiceModel.Dynamic.Test.ServiceModel.Cache.ICache CacheIstance
        {
            get
            {
                return _cache ?? (_cache = System.ServiceModel.Dynamic.Test.ServiceModel.Cache.CacheFactory.Instance.Get(CacheType));
            }
        }

        

        #endregion
    }
}
